#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import logging
import xlsxwriter
import StringIO
from datetime import datetime, date, time
from time import strftime, gmtime
from zipfile import ZipFile

from rest_framework import viewsets, status
from rest_framework.response import Response
from rest_framework import generics
from rest_framework.decorators import detail_route, list_route
from rest_framework.serializers import ValidationError
from rest_framework.exceptions import PermissionDenied

from django.db import transaction
from django.db.models import ProtectedError, Min, Q, Max
from django.conf import settings
from django.http import HttpResponse, HttpResponseRedirect, Http404, \
    StreamingHttpResponse
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.servers.basehttp import FileWrapper
from django.utils.dateparse import parse_datetime

from apimas.drf.mixins import HookMixin

from apella.models import InstitutionManager, Position, Department, \
    Candidacy, ApellaFile, ElectorParticipation, Candidate, \
    Professor as ProfessorModel, UserApplication, ApellaUser, \
    Registry, RegistryMembership, ForeignDiplomaDetails
from apella.loader import adapter
from apella.common import FILE_KIND_TO_FIELD, RANKS_EL, POSITION_STATES_EL
from apella import auth_hooks
from apella.serializers.position import link_files, \
    upgrade_candidate_to_professor, upgrade_candidates_to_professors
from apella.emails import send_user_email, send_emails_file, \
    send_emails_members_change, send_disable_professor_emails, \
    send_release_shibboleth_email
from apella.util import urljoin, safe_path_join, otz, move_to_timezone, \
    write_row, strip_timezone
from apella.serials import get_serial
from apella.validators import validate_candidate_files, \
    validate_discipline_registry_activation


logger = logging.getLogger(__name__)

DATE_EXPORT_FORMAT = "%Y%m%d%H%M"

def get_registry_type_el(registry_type):
    return'εσωτερικό' if registry_type == 'internal' else 'εξωτερικό'


def get_p_state(p):
    if p.state == 'posted' and p.starts_at \
            and p.starts_at > datetime.utcnow():
        p_state = 'Ενταγμένη'
    elif p.state == 'posted' and p.ends_at \
            and p.ends_at < datetime.utcnow():
        p_state = 'Κλειστή'
    elif p.state == 'posted':
        p_state = 'Ανοιχτή'
    else:
        p_state = POSITION_STATES_EL.get(p.state)
    return p_state

def get_p_rank(p):
    rank_el = ""
    if p.rank:
        rank_el = RANKS_EL.get(p.rank)
    return rank_el

def get_r_state(r):
    r_state = 'Πρόχειρο'
    if r.deactivated_at:
        r_state = 'Ανενεργό'
    elif r.is_active:
        r_state = 'Ενεργό'
    return r_state

class DestroyProtectedObject(viewsets.ModelViewSet):
    def destroy(self, request, *args, **kwargs):
        try:
            instance = self.get_object()
            self.perform_destroy(instance)
            return Response(status=status.HTTP_204_NO_CONTENT)
        except ProtectedError:
            return Response(status=status.HTTP_403_FORBIDDEN)


class Professor(object):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "professors_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Professors')

        user = request.user
        if user.is_manager() or user.is_ministry() or user.is_yyet():
            report_type = '1'
        elif user.is_helpdeskadmin():
            report_type = '2'
        else:
            return Response(status=status.HTTP_403_FORBIDDEN)


        if report_type == '1':
            fields = ['Κωδικός Χρήστη', 'Όνομα', 'Επώνυμο',
                'Φορέας', 'Σχολή', 'Τμήμα/Ινστιτούτο', 'Βαθμίδα', 'URL Βιογραφικού',
                'Γνωστικό Αντικείμενο', 'ΦΕΚ Διορισμού', 'Σε αναστολή',
                'Κατηγορία Χρήστη', 'Παρ. 1, Άρ. 145, Ν. 4957/2022']
        elif report_type == '2':
            fields = ['Κωδικός Χρήστη', 'Παλιός Κωδικός Χρήστη', 'Όνομα (el)',
                'Επώνυμο (el)', 'Πατρώνυμο (el)', 'Όνομα (en)', 'Επώνυμο (en)',
                'Πατρώνυμο (en)', 'Email', 'Κινητό Τηλέφωνο',
                'Σταθερό Τηλέφωνο', 'Αρ. Ταυτότητας',
                'Ημ/νία Δημιουργίας Λογαριασμού', 'login',
                'Ενεργός', 'Ενεργοποιήθηκε στις', 'Ενεργοποίηση Email',
                'Ενεργοποίηση Email στις',
                'Κατάσταση Προφίλ',
                'Ημ/νία Τελευταίας Αλλαγής Κατάστασης Προφίλ',
                'Κατηγορία Χρήστη', 'Φορέας', 'Κωδικός Φορέα',
                'Τμήμα/Ινστιτούτο', 'Κωδικός Τμήματος/Ινστιτούτου', 'Βαθμίδα',
                'Υπάρχει URL Βιογραφικού', 'URL Βιογραφικού',
                'Γνωστικό Αντικείμενο', 'Υπάρχει Γνωστικό Αντικείμενο στο ΦΕΚ',
                'ΦΕΚ Διορισμού', 'Ομιλεί Ελληνικά',
                'Έχει αποδεχτεί τους όρους συμμετοχής',
                'Shibboleth id',
                'Σε αναστολή',
                'Ακαδημαϊκός',
                'Παρ. 1, Άρ. 145, Ν. 4957/2022']

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        queryset = self.get_queryset().select_related(
                'user__first_name__el', 'user__last_name__el',
                'user__father_name__el', 'user', 'institution',
                'institution__title__el', 'department__title__el',
                'department', 'department__school__title__el')

        if report_type == '1':
            queryset = queryset.filter(is_verified=True, user__is_active=True)

        if user.is_manager() or user.is_yyet() or user.is_ministry():
            is_new_academy_member = Q(is_academy_member=True) & \
                Q(is_existing_academy_member=False)
            queryset = queryset.exclude(is_new_academy_member)

        i = 1
        for p in queryset:
            institution_id = '-'
            try:
                institution = p.institution.title.el
                institution_id = p.institution.id
            except AttributeError:
                institution = p.institution_freetext

            department_id = '-'
            try:
                department = p.department.title.el
                department_id = p.department.id
            except AttributeError:
                department = ''

            try:
                school = p.department.school.title.el
            except AttributeError:
                school = ''

            category = ''
            if p.is_professor and not p.is_foreign:
                category = 'Καθηγητής Ημεδαπής'
            elif p.is_professor and p.is_foreign:
                category = 'Καθηγητής Αλλοδαπής'
            elif not p.is_professor and not p.is_foreign:
                category = 'Ερευνητής Ημεδαπής'
            else:
                category = 'Ερευνητής Αλλοδαπής'

            rank_el = ""
            if p.rank:
                rank_el = RANKS_EL.get(p.rank)

            if report_type == '1':
                row = [
                    p.user.id,
                    p.user.first_name.el,
                    p.user.last_name.el,
                    institution,
                    school,
                    department,
                    rank_el and rank_el.decode('utf-8'),
                    p.cv_url,
                    p.discipline_text,
                    p.fek,
                    'ΝΑΙ'.decode('utf-8') \
                        if p.is_disabled else 'ΟΧΙ'.decode('utf-8'),
                    category.decode('utf-8'),
                    'ΝΑΙ'.decode('utf-8') \
                        if p.ad_personam else 'ΟΧΙ'.decode('utf-8'),
                ]
            elif report_type == '2':
                profile_state = ''
                profile_last_changed_at = '-'
                if p.is_verified:
                    profile_state = 'Πιστοποιημένος'
                    profile_last_changed_at = \
                        move_to_timezone(p.verified_at, otz)
                elif p.is_rejected:
                    profile_state = 'Απορριφθείς'
                elif p.verification_pending:
                    profile_state = 'Αναμονή Πιστοποίησης'
                    profile_last_changed_at = \
                        move_to_timezone(p.verification_request, otz)
                elif not p.verification_pending and not p.is_rejected \
                        and not p.is_verified and p.changes_request:
                    profile_state = 'Ζητήθηκαν αλλαγές'
                    profile_last_changed_at = \
                        move_to_timezone(p.changes_request, otz)

                discipline_in_fek_verbose = 'ΟΧΙ'.decode('utf-8')
                if p.is_foreign:
                    discipline_in_fek_verbose = ''
                elif p.discipline_in_fek:
                    discipline_in_fek_verbose = 'ΝΑΙ'.decode('utf-8')

                row = [
                    p.user.id,
                    p.user.old_user_id,
                    p.user.first_name.el,
                    p.user.last_name.el,
                    p.user.father_name.el,
                    p.user.first_name.en,
                    p.user.last_name.en,
                    p.user.father_name.en,
                    p.user.email,
                    p.user.mobile_phone_number,
                    p.user.home_phone_number,
                    p.user.id_passport,
                    str(move_to_timezone(p.user.date_joined, otz)),
                    p.user.login_method,
                    p.user.is_active,
                    str(move_to_timezone(p.user.activated_at, otz)),
                    p.user.email_verified,
                    str(move_to_timezone(p.user.email_verified_at, otz)),
                    profile_state.decode('utf-8'),
                    str(profile_last_changed_at),
                    category.decode('utf-8'),
                    institution,
                    institution_id,
                    department,
                    department_id,
                    rank_el and rank_el.decode('utf-8'),
                    'ΝΑΙ'.decode('utf-8') \
                        if p.cv_url else 'ΟΧΙ'.decode('utf-8'),
                    p.cv_url,
                    p.discipline_text,
                    discipline_in_fek_verbose,
                    p.fek,
                    'ΝΑΙ'.decode('utf-8') \
                        if p.speaks_greek else 'ΟΧΙ'.decode('utf-8'),
                    'ΝΑΙ'.decode('utf-8') \
                        if p.user.has_accepted_terms \
                        else 'ΟΧΙ'.decode('utf-8'),
                    p.user.shibboleth_id,
                    'ΝΑΙ'.decode('utf-8') \
                        if p.is_disabled else 'ΟΧΙ'.decode('utf-8'),
                    'ΝΑΙ'.decode('utf-8') \
                        if p.is_academy_member and not p.is_existing_academy_member
                        else 'ΟΧΙ'.decode('utf-8'),
                    'ΝΑΙ'.decode('utf-8') \
                        if p.ad_personam else 'ΟΧΙ'.decode('utf-8'),
                    #p.active_elections
                ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

    def get_queryset(self):
        queryset = self.queryset
        leave_query = self.request.GET.get('on_leave')
        if leave_query:
            now = datetime.utcnow()
            queryset = queryset.filter(Q(leave_starts_at__lte=now) &
                                       Q(leave_ends_at__gt=now))

        is_new_academy_member = Q(is_academy_member=True) & \
            Q(is_existing_academy_member=False)
        exclude_new_academy = self.request.GET.get('is_new_academy_member')
        if exclude_new_academy:
            queryset = queryset.exclude(is_new_academy_member)

        if 'ordering' not in self.request.query_params:
            queryset = self.queryset.order_by('user__last_name__el')
        create_registry = self.request.query_params.get(
            'create_registry', None)
        if create_registry:
            queryset = queryset.exclude(rank='Lecturer'). \
                exclude(rank='Tenured Assistant Professor'). \
                exclude(is_disabled=True). \
                exclude(is_new_academy_member). \
                exclude(institution_id__in=settings.MILITARY_INSTITUTION_IDS)
            registry_id = self.request.query_params.get(
                'registry_id', None)
            if registry_id:
                try:
                    registry = Registry.objects.get(id=registry_id)
                    queryset = queryset.exclude(
                        registrymembership__registry=registry)
                    other_type = 'internal' if registry.type == 'external' \
                        else 'external'
                    other_registry = Registry.objects.get(
                        department=registry.department,
                        type=other_type)
                    queryset = queryset.exclude(
                        registrymembership__registry=other_registry)
                except Registry.DoesNotExist:
                    pass
        return queryset

    def set_professor_is_disabled(self, is_disabled, request):
        professor = self.get_object()
        if professor.is_disabled is is_disabled:
            return
        professor.is_disabled = is_disabled
        if is_disabled:
            professor.disabled_at = datetime.utcnow()
            professor.disabled_by_helpdesk = request.user.is_helpdesk()
        professor.save()
        logger.info(
            'user %s %s professor %r' %
            (request.user.username,
            'disabled' if is_disabled else 'enabled',
            professor.id))
        if is_disabled:
            send_disable_professor_emails(professor, is_disabled)

    @detail_route(methods=['post'])
    def disable_professor(self, request, pk=None):
        professor = self.get_object()
        try:
            self.set_professor_is_disabled(True, request)
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def enable_professor(self, request, pk=None):
        professor = self.get_object()
        try:
            self.set_professor_is_disabled(False, request)
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route()
    def participations(self, request, pk=None):
        professor = self.get_object()
        response = HttpResponse(content_type='application/ms-excel')
        filename = "participations_export_" + str(professor.user.pk) + "_" +\
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Participations')

        fields = [
            'κωδικός θέσης που συμμετέχει', 'id φορέα θέσης', 'φορέας θέσης',
            'id τμήματος θέσης', 'τμήμα θέσης', 'κατάσταση θέσης', 'βαθμίδα θέσης', 
            'ρόλος',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        # get participations as elector
        eps = ElectorParticipation.objects.filter(
                professor=professor,
                position__is_historical=False)
        for ep in eps:
            p = ep.position

            member_type = 'τακτικό' if ep.is_regular else 'αναπληρωματικό'

            registry_type = 'εξωτερικό μητρώο'
            if ep.is_internal:
                registry_type = 'εσωτερικό μητρώο'
            if ep.is_academy:
                registry_type = 'μέλος Ακαδημίας Αθηνών'

            elector_type = 'επιτροπής κρίσης'
            if ep.position.department.institution_category == 'Institution':
                elector_type = 'εκλεκτορικού σώματος'

           
            role = '%s μέλος %s (%s)' % \
                    (member_type, elector_type, registry_type)
            
            row = [
                ep.position.id,
                ep.position.department.institution.id,
                ep.position.department.institution.title.el,
                ep.position.department.id,
                ep.position.department.title.el,
                get_p_state(ep.position).decode('utf-8'),
                get_p_rank(ep.position).decode('utf-8'),
                role.decode('utf-8'),
            ]
            write_row(ws, row, i)
            i += 1

        # get participations as committee
        cpositions = professor.committee_duty.filter(is_historical=False)
        for p in cpositions:

            registry__type = 'εξωτερικό μητρώο'
            if RegistryMembership.objects.filter(
                professor=professor,
                registry__department=p.department,
                registry__type='internal').exists():
                registry_type = 'εσωτερικό μητρώο'
            role = 'εισηγητική επιτροπή (%s)' % (registry_type)

            row = [
                p.id,
                p.department.institution.id,
                p.department.institution.title.el,
                p.department.id,
                p.department.title.el,
                get_p_state(p).decode('utf-8'),
                get_p_rank(p).decode('utf-8'),
                role.decode('utf-8'),
            ]
            write_row(ws, row, i)
            i += 1


        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response




class AssistantList(generics.ListAPIView):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "assistants_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Assistants')

        fields = [
            'κωδικός χρήστη','όνομα βοηθού', 'επώνυμο βοηθού', 'username', 'email',
            'κατηγορία βοηθού', 'id φορέα', 'φορέας', 'κατηγορία φορέα',
            'id τμημάτων/ινστιτούτων', 'τμήματα/ινστιτούτα',
            'δημιουργία θέσεων', 'δημιουργία μητρώων', 
            'status πιστοποίησης', 'status λογαριασμού',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        assistants = self.get_queryset()
        for a in assistants:

            a_category = ''
            if a.is_secretary:
                a_category = 'χρήστης γραμματείας'
            else:
                if a.institution.category == 'Institution':
                    a_category = 'χρήστης ιδρύματος'
                else:
                    a_category = 'χρήστης ερευνητικού'

            profile_state = ''
            if a.is_verified:
                profile_state = 'Πιστοποιημένος'
            elif a.is_rejected:
                profile_state = 'Απορριφθείς'

            user_state = ''
            if a.user.is_active:
                user_state = 'Ενεργός'
            else:
                user_state = 'Ανενεργός'

            a_departments = a.departments.all()
            a_deps_ids_list = [str(d.id) for d in a_departments]
            a_deps_titles_list = [d.title.el for d in a_departments]
            a_deps_ids = ','.join(a_deps_ids_list)
            a_deps_titles = ','.join(a_deps_titles_list)
            
            row = [
                a.user.id,
                a.user.first_name.el,
                a.user.last_name.el,
                a.user.username,
                a.user.email,
                a_category.decode('utf-8'),
                a.institution.id,
                a.institution.title.el,
                a.institution.category,
                a_deps_ids,
                a_deps_titles,
                1 if a.can_create_positions else 0,
                1 if a.can_create_registries else 0,
                profile_state.decode('utf-8'),
                user_state.decode('utf-8'),
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

    def get_queryset(self):
        user = self.request.user
        if user.is_manager():
            institution_id = user.institutionmanager.institution_id
            return InstitutionManager.objects.filter(
                manager_role='assistant',
                institution__id=institution_id)
        elif user.is_yyet():
            return InstitutionManager.objects.filter(
                manager_role='assistant',
                institution__category='Research')
        elif user.is_ministry():
            return InstitutionManager.objects.filter(
                manager_role='assistant',
                institution__category='Institution')
        else:
            department = self.request.query_params.get('department', None)
            filters = Q()
            filters &= Q(manager_role='assistant')
            if department:
                filters &= Q(departments=department)

            return InstitutionManager.objects.filter(filters)


class InstitutionManagerList(generics.ListAPIView):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "institution_managers_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('InstitutionManagers')

        fields = [
            'Κωδικός χρήστη','Όνομα', 'Επώνυμο', 'username', 'email',
            'id φορέα', 'φορέας', 'κατηγορία φορέα',
            'ονοματεπώνυμο αναπληρωτή διαχειριστή', 'email αναπληρωτή διαχειριστή',
            'status πιστοποίησης', 'status λογαριασμού',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        institution_managers = self.get_queryset().filter(
                manager_role='institutionmanager')
        for m in institution_managers:

            sub_first_name = ''
            sub_last_name = ''
            if m.sub_first_name:
                sub_first_name = m.sub_first_name.el
            if m.sub_last_name:
                sub_last_name = m.sub_last_name.el
            sub_full_name = sub_first_name + ' ' + sub_last_name
            
            profile_state = ''
            if m.is_verified:
                profile_state = 'Πιστοποιημένος'
            elif m.is_rejected:
                profile_state = 'Απορριφθείς'

            user_state = ''
            if m.user.is_active:
                user_state = 'Ενεργός'
            else:
                user_state = 'Ανενεργός'

            row = [
                m.user.id,
                m.user.first_name.el,
                m.user.last_name.el,
                m.user.username,
                m.user.email,
                m.institution.id,
                m.institution.title.el,
                m.institution.category,
                sub_full_name,
                m.sub_email,
                profile_state.decode('utf-8'),
                user_state.decode('utf-8'),
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response


    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user

        if user.is_anonymous():
            return InstitutionManager.objects.none()

        if user.is_yyet():
            return InstitutionManager.objects.filter(
                institution__category='Research')

        if user.is_ministry():
            return InstitutionManager.objects.filter(
                institution__category='Institution')
        return queryset

elector_sets = {
    'electors_regular_internal': True,
    'electors_regular_external': True,
    'electors_sub_internal': False,
    'electors_sub_external': False,
    'electors_academy_regular': True,
    'electors_academy_sub': False
}

committee_sets = ['committee_internal', 'committee_external']


class PositionHookMixin(HookMixin):
    def preprocess_update(self):
        obj = self.unstash()
        position = obj.instance

        curr_position = Position.objects.get(id=position.id)
        if position.state == 'electing' and curr_position.state != 'revoked' \
                and curr_position.state != 'successful' and \
                curr_position.state != 'failed' and \
                curr_position.state != 'cancelled':

            eps = ElectorParticipation.objects.filter(position=position)
            old_participations = [old_el_pa for old_el_pa in eps.all()]
            eps.all().delete()

            new_participations = []
            for elector_set_key, is_regular in elector_sets.items():
                if elector_set_key in obj.validated_data:
                    is_academy = "academy" in elector_set_key
                    for elector in obj.validated_data[elector_set_key]:
                        ep = ElectorParticipation.objects.create(
                            professor=elector,
                            position=position,
                            is_regular=is_regular,
                            is_internal=elector_set_key.endswith('internal'),
                            is_academy=is_academy)

                        new_participations.append(ep)

            send_emails_members_change(
                position, 'electors', {'e': old_participations},
                {'e': new_participations})

            c = {'committee': []}
            for com_set in committee_sets:
                if obj.validated_data and com_set in obj.validated_data:
                    committee = obj.validated_data[com_set]
                    if committee:
                        for professor in committee:
                            c['committee'].append(professor)
            self.stash(extra=c)

class PositionMixin(object):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        position_for = self.request.GET.get('position_for', '')
        filename = "positions_export_" + position_for + '_'+ \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Positions')

        fields = ['Κωδικός Θέσης', 'Παλιός Κωδικός Θέσης',
            'Τίτλος Θέσης', 'Φορέας', 'Σχολή', 'Τμήμα/Ινστιτούτο', 'Βαθμίδα',
            'Περιγραφή', 'Γνωστικό Αντικείμενο', 'Θεματική Περιοχή',
            'Θέμα', 'Σχετιζόμενες Θέσεις', 'Φ.Ε.Κ.', 'Ημ/νία Φ.Ε.Κ.',
            'Κατάσταση Θέσης', 'Ημ/νία Έναρξης Υποβολών',
            'Ημ/νία Λήξης Υποβολών',
            'Ημ/νία Σύγκλησης του Εκλεκτορικού Σώματος για τον ορισμό της Εισηγητικής Επιτροπής',
            'Ημ/νία Σύγκλησης του Εκλεκτορικού Σώματος/Επιτροπής Κρίσης για Επιλογή',
            'Εκλεγείς', 'Δεύτερος Καταλληλότερος Υποψήφιος',
            'Φ.Ε.Κ. Διορισμού']

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        queryset = self.get_queryset().select_related(
            'elected__first_name__el', 'elected__last_name__el',
            'second_best__first_name__el', 'second_best__last_name__el',
            'department__title__el', 'department__school__title__el',
            'department__institution__title__el', 'subject__title__el',
            'subject_area__title__el')

        for p in queryset:
            elected_full_name = ''
            if p.elected:
                elected_full_name = \
                    p.elected.first_name.el + ' ' + \
                    p.elected.last_name.el

            second_best_full_name = ''
            if p.second_best:
                second_best_full_name = \
                    p.second_best.first_name.el + ' ' + \
                    p.second_best.last_name.el

            rank_el = ""
            if p.rank:
                rank_el = RANKS_EL.get(p.rank)

            if p.state == 'posted' and p.starts_at \
                    and p.starts_at > datetime.utcnow():
                p_state = 'Ενταγμένη'
            elif p.state == 'posted' and p.ends_at \
                    and p.ends_at < datetime.utcnow():
                p_state = 'Κλειστή'
            elif p.state == 'posted':
                p_state = 'Ανοιχτή'
            else:
                p_state = POSITION_STATES_EL.get(p.state)

            p_related = ""
            if p.related_positions.count() > 0:
                p_related = ','.join(str(rp.code) for rp in
                        p.related_positions.all())

            row = [
                p.code,
                p.old_code,
                p.title,
                p.department.institution.title.el,
                p.department.school and p.department.school.title.el,
                p.department.title.el,
                rank_el and rank_el.decode('utf-8'),
                p.description,
                p.discipline,
                p.subject_area.title.el,
                p.subject.title.el,
                p_related,
                p.fek,
                p.fek_posted_at,
                p_state and p_state.decode('utf-8'),
                p.starts_at and str(move_to_timezone(p.starts_at, otz).date()),
                p.ends_at and str(move_to_timezone(p.ends_at, otz).date()),
                p.electors_meeting_to_set_committee_date,
                p.electors_meeting_date,
                elected_full_name,
                second_best_full_name,
                p.nomination_act_fek
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

    @detail_route()
    def history(self, request, pk=None):
        position = self.get_object()
        position_states = Position.objects.filter(code=position.code). \
            order_by('-updated_at')
        serializer = self.get_serializer(position_states, many=True)
        return Response(serializer.data)

    @detail_route(methods=['post'])
    def edit_fields(self, request, pk=None):
        p = self.get_object()
        electors_meeting_date = request.data['electors_meeting_date']
        if electors_meeting_date:
            electors_meeting_date = strip_timezone(
                parse_datetime(electors_meeting_date))
        else:
            electors_meeting_date = None
        p.electors_meeting_date = electors_meeting_date

        electors_meeting_to_set_committee_date = request.data[
            'electors_meeting_to_set_committee_date']
        if electors_meeting_to_set_committee_date:
            electors_meeting_to_set_committee_date = strip_timezone(
                parse_datetime(electors_meeting_to_set_committee_date))
        else:
            electors_meeting_to_set_committee_date = None
        p.electors_meeting_to_set_committee_date = electors_meeting_to_set_committee_date

        department_dep_number = request.data['department_dep_number']
        if department_dep_number:
            p.department_dep_number = department_dep_number

        p.save()
        return Response(request.data, status=status.HTTP_200_OK)

    def get_queryset(self):
        now = datetime.utcnow()
        queryset = self.queryset
        user = self.request.user

        if 'pk' in self.kwargs:
            queryset = queryset.filter(id=self.kwargs['pk'])
        else:
            queryset = queryset.filter(is_historical=False)

        if user.is_authenticated():
            if user.is_institutionmanager():
                queryset = queryset.filter(
                    department__in=user.institutionmanager.
                    institution.department_set.all())
            elif user.is_assistant():
                queryset = queryset.filter(
                    department__in=user.institutionmanager.
                    departments.all())
            # query results for ministry only show institutional positions
            elif user.is_ministry():
                queryset = queryset.exclude(position_type='research'). \
                        exclude(position_type='tenure_research'). \
                        exclude(position_type='tenure_scientist')
            elif user.is_professor():
                # if honorable member of the academy just show positions they
                # participate as  electors
                if user.professor.is_academy_member and not \
                        user.professor.is_existing_academy_member:
                    queryset = queryset.filter(
                        Q(electors=user.professor.id))

                # if all other professor cases
                else:
                    position_codes = list(Candidacy.objects.filter(
                        candidate=user).values_list('position__code', flat=True))
                    # if professor is not researcher show the positions of their
                    # department, 

                    if user.professor.department and \
                        user.professor.department.institution.category == 'Institution':
                        department_position_codes = list(Position.objects.filter(
                            department=user.professor.department).
                            values_list('code', flat=True))
                        position_codes += department_position_codes

                    position_codes += list(
                        user.userapplication_set.values_list(
                        'position__code', flat=True))
                    queryset = queryset.filter(
                        Q(state='posted', ends_at__gte=now) |
                        Q(code__in=position_codes) |
                        Q(committee=user.professor.id) |
                        Q(electors=user.professor.id))
            elif user.is_candidate():
                position_codes = Candidacy.objects.filter(
                    candidate=user).values_list('position__code', flat=True)
                queryset = queryset.filter(
                    Q(state='posted', ends_at__gte=now) |
                    Q(code__in=position_codes))

            state_query = self.request.GET.get('state_expanded')
            if state_query:
                now = datetime.utcnow()
                if state_query in ('electing', 'successful',
                                   'failed', 'cancelled', 'revoked'):
                    queryset = queryset.filter(Q(state=state_query))

                elif state_query == 'posted':
                    queryset = queryset.filter(Q(state='posted') &
                                               Q(starts_at__gte=now))

                elif state_query == 'open':
                    queryset = queryset.filter(Q(state='posted') &
                                               Q(starts_at__lte=now) &
                                               Q(ends_at__gt=now))
                elif state_query == 'closed':
                    queryset = queryset.filter(Q(state='posted') &
                                               Q(ends_at__lte=now))

                elif state_query == 'before_closed':
                    queryset = queryset.filter(
                        (Q(state='posted') &
                        Q(ends_at__gt=now)) |
                        Q(ends_at=None))

            position_for = self.request.GET.get('position_for')
            if position_for:
                if position_for == 'research':
                    queryset = queryset.filter(
                            Q(position_type='research') |
                            Q(position_type='tenure_scientist') |
                            Q(position_type='tenure_research'))
                else:
                    queryset = queryset.exclude(position_type='research'). \
                            exclude(position_type='tenure_research'). \
                            exclude(position_type='tenure_scientist')

            queryset = queryset.distinct()

            # Handle case when positions are queried by professor participation 
            # as elector
            if 'elector_participation' in self.request.query_params and \
                    'electors' in self.request.query_params:
                elector_participation = self.request.query_params['elector_participation']
                electors = self.request.query_params['electors']

                # default values
                is_regular=False
                is_academy=False
                is_internal=False

                if 'regular' in elector_participation: 
                    is_regular=True

                if 'internal' in elector_participation:
                    is_internal=True

                if 'academy' in elector_participation:
                    is_academy=True


                eps = ElectorParticipation.objects.filter(professor=electors,\
                        is_regular=is_regular, 
                        is_academy=is_academy,
                        is_internal=is_internal)
                pks = [e.position.pk for e in eps]
                queryset = queryset.filter(pk__in=pks)

            # Handle case when positions are queried by professor participation 
            # as committee
            if 'committee_participation' in self.request.query_params and \
                    'committee' in self.request.query_params:
                committee_participation = self.request.query_params['committee_participation']
                committee = self.request.query_params['committee']

                registry_type='internal' if committee_participation=='internal' else 'external'
                p_ids = []
                for p in queryset:
                    if RegistryMembership.objects.filter(
                            professor=committee,
                            registry__department=p.department,
                            registry__type=registry_type).exists():
                        p_ids.append(p.pk)

                queryset = queryset.filter(pk__in=p_ids)
        return queryset

    def update(self, request, *args, **kwargs):
        position = self.get_object()
        code = position.code
        if code.split(settings.POSITION_CODE_PREFIX)[1] != \
                str(position.id):
            return Response(status=status.HTTP_403_FORBIDDEN)
        return super(PositionMixin, self).update(request, *args, **kwargs)


class CandidacyList(object):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "candidacies_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Candidacies')

        fields = [
            'κωδικός υποψηφιότητας', 'κωδικός χρήστη', 'όνομα', 'επώνυμο',
            'ημερομηνία υποβολής', 'ημερομηνία τελευταίας ενημέρωσης',
            'κατάσταση υποψηφιότητας',
            'id φορέα', 'φορέας', 'id τμήματος/ινστιτούτου',
            'τμήμα/ινστιτούτο', 'κατηγορία φορέα', 
            'id θέσης', 'τίτλος θέσης', 
            'βαθμίδα',
            'γνωστικό αντικείμενο θέσης', 'κατάσταση θέσης',
            'ταυτότητα ή διαβατήριο', 'βιογραφικό', 'αντίγραφα πτυχίων',
            'αρχεία δημοσιεύσεων', 'υπόμνημα επιστημονικών δημοσιεύσεων',
            'έκθεση αυτοαξιολόγησης', 'συμπληρωματικά Έγγραφα',
            'υπόμνημα προς το εκλεκτορικό σώμα', 'υπόμνημα προς τον Πρύτανη',
            'συγχρονισμός ταυτότητας'
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        candidacies = self.get_queryset().select_related(
                'position__department__institution__id',
                'position__department__institution__title__el',
                'position__department__id',
                'position__department__title__el',
                'position__department__institution__category',
                'position__id',
                'position__title',
                'position__rank',
                'position__discipline',
                'position__state',
                'position__starts_at', 'position__ends_at',
                'candidate__id',
                'candidate__first_name__el', 'candidate__last_name__el',
                )
        for c in candidacies:

            candidacy_state = ''
            if c.state == 'draft':
                candidacy_state = 'Πρόχειρο'
            elif c.state == 'posted':
                candidacy_state = 'Υποβληθείσα'
            elif c.state == 'cancelled':
                candidacy_state = 'Αποσυρθείσα'

            p = c.position
            if p.state == 'posted' and p.starts_at \
                    and p.starts_at > datetime.utcnow():
                p_state = 'Ενταγμένη'
            elif p.state == 'posted' and p.ends_at \
                    and p.ends_at < datetime.utcnow():
                p_state = 'Κλειστή'
            elif p.state == 'posted':
                p_state = 'Ανοιχτή'
            else:
                p_state = POSITION_STATES_EL.get(p.state)

            synced_id_file = 0
            try:
                p = ProfessorModel.objects.get(user__id=c.candidate.id)
                if p.synced_id_file:
                    synced_id_file = 1
            except ProfessorModel.DoesNotExist:
                try:
                    candidate = Candidate.objects.get(user__id = c.candidate.id)
                    if candidate.synced_id_file:
                        synced_id_file = 1
                except Candidate.DoesNotExist:
                    pass

            rank_el = ""
            if c.position.rank:
                rank_el = RANKS_EL.get(c.position.rank)


            row = [
                c.id,
                c.candidate.id,
                c.candidate.first_name.el,
                c.candidate.last_name.el,
                c.submitted_at.strftime("%Y-%m-%dT%H:%M"),
                c.updated_at.strftime("%Y-%m-%dT%H:%M"),
                candidacy_state.decode('utf-8'),
                c.position.department.institution.id,
                c.position.department.institution.title.el,
                c.position.department.id,
                c.position.department.title.el,
                c.position.department.institution_category,
                c.position.id,
                c.position.title,
                rank_el.decode('utf-8'),
                c.position.discipline,
                p_state.decode('utf-8'),
                1 if c.id_file else 0,
                1 if c.cv else 0,
                c.diplomas.count(),
                c.publications.count(),
                1 if c.pubs_note else 0,
                1 if c.self_evaluation_report else 0,
                c.attachment_files.count(),
                1 if c.statement_file else 0,
                1 if c.statement_in_opposition_file else 0,
                synced_id_file,
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response


    @detail_route()
    def history(self, request, pk=None):
        candidacy = self.get_object()
        candidacy_states = Candidacy.objects.filter(code=candidacy.code)
        serializer = self.get_serializer(candidacy_states, many=True)
        return Response(serializer.data)

    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user
        if user.is_institutionmanager():
            institution_ids = InstitutionManager.objects.filter(user=user). \
                values_list('institution', flat=True)
            departments = Department.objects.filter(
                institution_id__in=institution_ids)
            positions = Position.objects.filter(department__in=departments)
            queryset = queryset.filter(position__in=positions)
        elif user.is_yyet():
            departments = Department.objects.filter(
                institution__category='Research')
            positions = Position.objects.filter(department__in=departments)
            queryset = queryset.filter(position__in=positions)
        elif user.is_ministry():
            departments = Department.objects.filter(
                institution__category='Institution')
            positions = Position.objects.filter(department__in=departments)
            queryset = queryset.filter(position__in=positions)
        elif user.is_assistant():
            departments = user.institutionmanager.departments.all()
            positions = Position.objects.filter(department__in=departments)
            queryset = queryset.filter(position__in=positions)
        if 'pk' in self.kwargs:
            queryset = queryset.filter(id=self.kwargs['pk'])
        else:
            ids = queryset.values('code').annotate(Min('id')). \
                values_list('id__min', flat=True)
            queryset = queryset.filter(id__in=ids)
            if 'latest' in self.request.query_params:
                q2 = queryset.values('candidate', 'position'). \
                    annotate(Max('id')).values_list('id__max', flat=True)
                queryset = queryset.filter(id__in=q2)
        if not user.is_helpdesk():
            queryset = queryset.exclude(state='draft')

        position_state_query = self.request.GET.get('position_state_expanded')
        if position_state_query:
            now = datetime.utcnow()
            if position_state_query in ('electing', 'successful',
                               'failed', 'cancelled', 'revoked'):
                queryset = queryset.filter(Q(position__state=position_state_query))

            elif position_state_query == 'posted':
                queryset = queryset.filter(Q(position__state='posted') &
                                           Q(position__starts_at__gte=now))

            elif position_state_query == 'open':
                queryset = queryset.filter(Q(position__state='posted') &
                                           Q(position__starts_at__lte=now) &
                                           Q(position__ends_at__gt=now))
            elif position_state_query == 'closed':
                queryset = queryset.filter(Q(position__state='posted') &
                                           Q(position__ends_at__lte=now))

            elif position_state_query == 'before_closed':
                queryset = queryset.filter(
                    (Q(position__state='posted') &
                    Q(position__ends_at__gt=now)) |
                    Q(position__ends_at=None))

        queryset = queryset.distinct()
        return queryset

    def update(self, request, pk=None):
        candidacy = self.get_object()
        code = candidacy.code
        if code != str(candidacy.id):
            return Response(status=status.HTTP_403_FORBIDDEN)
        return super(CandidacyList, self).update(request, pk=None)

    @list_route(methods=['POST'])
    def upgrade_role(self, request, pk=None):

        file_upload = request.FILES.get('file_upload', None)
        if not file_upload:
            return Response('no.csv.file', status=status.HTTP_400_BAD_REQUEST)

        owner = request.user
        try:
            errors, success = \
                upgrade_candidates_to_professors(file_upload, owner)
        except ValidationError as ve:
            return Response(
                {'errors': ve.detail}, status=status.HTTP_400_BAD_REQUEST)

        if not errors:
            return Response(
                {'success': success}, status=status.HTTP_200_OK)
        else:
            return Response(
                {'errors': errors, 'success': success},
                    status=status.HTTP_409_CONFLICT)

    @detail_route()
    def download_zip(self, request, pk=None):
        candidacy = self.get_object()

        zip_subdir = "candidacy_%r_files_" % candidacy.id
        now = datetime.utcnow()
        local_now = move_to_timezone(now, otz)
        zip_subdir += local_now.strftime("%Y%m%d_%H%M%S")
        zip_filename = "%s.zip" % zip_subdir

        s = StringIO.StringIO()
        with ZipFile(s, "w") as zf:
            # write cv file
            zf.write(
                candidacy.cv.file_content.path,
                os.path.join(zip_subdir, "cv", candidacy.cv.file_name))
            # write diplomas
            for d in candidacy.diplomas.all():
                zf.write(
                    d.file_content.path,
                    os.path.join(zip_subdir, "diplomas", d.file_name))
            # write diplomas
            for fd in candidacy.foreign_diplomas.all():
                zf.write(
                    fd.file_content.path,
                    os.path.join(zip_subdir, "foreign_diplomas", fd.file_name))
                fdd = fd.foreign_diploma_details
                if fdd:
                    for extra_file in fdd.foreign_diploma_files.all():
                        zf.write(
                            extra_file.file_content.path,
                            os.path.join(
                                zip_subdir, "foreign_diplomas", extra_file.file_name))

            # write publications
            for p in candidacy.publications.all():
                zf.write(
                    p.file_content.path,
                    os.path.join(zip_subdir, "publications", p.file_name))
            # write self evaluation report
            if candidacy.self_evaluation_report:
                zf.write(
                    candidacy.self_evaluation_report.file_content.path,
                    os.path.join(
                        zip_subdir,
                        "self_evaluation_report",
                        candidacy.self_evaluation_report.file_name))
            # write attachment files
            if candidacy.attachment_files.count() > 0:
                for att in candidacy.attachment_files.all():
                    zf.write(
                        att.file_content.path,
                        os.path.join(zip_subdir, "attachment_files", att.file_name))
            # write statement file
            if candidacy.statement_file:
                zf.write(
                    candidacy.statement_file.file_content.path,
                    os.path.join(
                        zip_subdir,
                        "statement_file",
                        candidacy.statement_file.file_name))
            # write statement in opposition file
            if candidacy.statement_in_opposition_file:
                zf.write(
                    candidacy.statement_in_opposition_file.file_content.path,
                    os.path.join(
                        zip_subdir,
                        "statement_in_opposition_file",
                        candidacy.statement_in_opposition_file.file_name))
            # write id file
            if candidacy.id_file:
                zf.write(
                    candidacy.id_file.file_content.path,
                    os.path.join(
                        zip_subdir,
                        "id_file",
                        candidacy.id_file.file_name))
            # write publications note file
            if candidacy.pubs_note:
                zf.write(
                    candidacy.pubs_note.file_content.path,
                    os.path.join(
                        zip_subdir,
                        "publications_note",
                        candidacy.pubs_note.file_name))

        response = HttpResponse(content_type='application/zip')
        response['Content-Disposition'] = 'attachment; filename=%s' % zip_filename
        response.write(s.getvalue())
        return response


class RegistriesList(viewsets.GenericViewSet):
    @list_route()
    def report_summary(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "registries_summary_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Registries')

        fields = ['id μητρώου', 'id φορέα', 'φορέας', 'σχολή',
                  'id τμήματος/ινστιτούτου', 'τμήμα/ινστιτούτο', 'κατηγορία φορέα',
                  'είδος μητρώου', 'αριθμός μελών']

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        queryset = self.get_queryset()

        for r in queryset:

            try:
                school = r.department.school.title.el
            except AttributeError:
                school = ''

            row = [
                r.id,
                r.department.institution.id,
                r.department.institution.title.el,
                school,
                r.department.id,
                r.department.title.el,
                'Ίδρυμα'.decode('utf-8') \
                    if r.department.institution.category == 'Institution' \
                    else 'Ερευνητικό κέντρο'.decode('utf-8'),
                'Εσωτερικό'.decode('utf-8') \
                    if r.type == 'internal' \
                    else 'Εξωτερικό'.decode('utf-8'),
                r.members_count
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response


    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "registries_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Registries')

        fields = ['Κωδικός Μητρώου', 'Φορέας Μητρώου', 'Τμήμα/Ινστιτούτο Μητρώου',
                  'Είδος Μητρώου', 'Κωδικός Χρήστη', 'Όνομα', 'Επώνυμο',
                  'Κατηγορία Χρήστη', 'Φορέας Χρήστη', 'Σχολή Χρήστη',
                  'Τμήμα/Ινστιτούτο Χρήστη', 'ΦΕΚ Διορισμού', 'Γνωστικό Αντικείμενο',
                  'Βαθμίδα']

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        queryset = self.get_queryset()
        for r in self.filtered_queryset(queryset):
            memberships = RegistryMembership.objects.filter(
                registry=r).select_related(
                'professor__institution__title__el',
                'professor__department__title__el',
                'professor__department__school__title__el',
                'professor__user__first_name__el',
                'professor__user__last_name__el',
                'registry__department__institution',
                'registry__department__institution__title__el',
                'registry__department__school__title__el',
                'registry__department__title__el',
                'registry__department').all()
            for m in memberships:
                try:
                    institution = m.professor.institution.title.el
                except AttributeError:
                    institution = m.professor.institution_freetext

                try:
                    department = m.professor.department.title.el
                except AttributeError:
                    department = ''

                try:
                    school = m.professor.department.school.title.el
                except AttributeError:
                    school = ''

                rank_el = ""
                if m.professor.rank:
                    rank_el = RANKS_EL.get(m.professor.rank)

                row = [
                    r.id,
                    r.department.institution.title.el,
                    r.department.title.el,
                    'Εσωτερικό'.decode('utf-8') \
                        if r.type == 'internal' \
                        else 'Εξωτερικό'.decode('utf-8'),
                    m.professor.user.id,
                    m.professor.user.first_name.el,
                    m.professor.user.last_name.el,
                    'Αλλοδαπής'.decode('utf-8') \
                        if m.professor.is_foreign else 'Ημεδαπής'.decode('utf-8'),
                    institution,
                    school,
                    department,
                    m.professor.fek,
                    m.professor.discipline_text,
                    rank_el and rank_el.decode('utf-8')
                ]
                write_row(ws, row, i)
                i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user
        if 'ordering' not in self.request.query_params:
            queryset = queryset.order_by(
                    'department__institution__title__el',
                    'department__title__el')
        else:
            ordering = self.request.query_params['ordering']
            queryset = queryset.order_by(ordering)
        if 'active_participation' in self.request.query_params and user.is_professor():
            registries = user.professor.registries
            queryset = queryset.filter(id__in=registries)
        return queryset

    def filtered_queryset(self, queryset):
        user = self.request.user
        if user.is_helpdesk():
            return queryset
        elif user.is_assistant():
            assistantDepartments = user.institutionmanager.departments.all()
            return queryset.filter(department__in=assistantDepartments)
        elif user.is_manager():
            managerInstitution = user.institutionmanager.institution
            return queryset.filter(department__institution=managerInstitution)
        else:
            return []


class DepartmentsList(viewsets.GenericViewSet):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "departments_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Departments')

        fields = [
            'id φορέα', 'el φορέα', ' en φορέα', 'κατηγορία φορέα', 
            'id τμήματος/ινστιτούτου', 'el τμήματος/ινστιτούτου', 
            'en τμήματος/ινστιτούτου', 'has_shibboleth',
            'idp', 'schac_home_organization', 'mineduid',
            'el σχολής', 'en σχολής',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        departments = self.get_queryset()
        for d in departments:
            
            if d.school: 
                school_el = d.school.title.el
                school_en = d.school.title.en
            else: 
                school_el = ''
                school_en = ''

            row = [
                d.institution.id,
                d.institution.title.el,
                d.institution.title.en,
                d.institution.category,
                d.id,
                d.title.el,
                d.title.en,
                d.institution.has_shibboleth,
                d.institution.idp,
                d.institution.schac_home_organization,
                d.mineduid,
                school_el,
                school_en,
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response


class SubjectsList(viewsets.GenericViewSet):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "subjects_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Subjects')

        fields = [
            'subject area id', 'subject area el', 'subject area en', 'subject id',
             'subject el', 'subject en', 'positions',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        subjects = self.get_queryset()
        for s in subjects:
            row = [
                s.area.id,
                s.area.title.el,
                s.area.title.en,
                s.id,
                s.title.el,
                s.title.en,
                s.position_set.all().count(),
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response


def os_remove_file_and_log(obj, username):
    try:
        f = safe_path_join(settings.MEDIA_ROOT, obj.file_content.name)
        os.remove(f)
        logger.info(
            'user %s removed file: %s, path: %s source: %s, source_id: %s' %
            (username,
            obj.file_name,
            obj.file_content.path,
            obj.source,
            obj.source_id,))
    except OSError:
        pass

USE_X_SEND_FILE = getattr(settings, 'USE_X_SEND_FILE', False)

class FilesViewSet(viewsets.ModelViewSet):

    @detail_route(methods=['head'], url_path='download')
    def download_head(self, request, pk=None):
        response = HttpResponse(content_type='application/force-download')
        assert request.method == 'HEAD'
        user = request.user
        file = self.get_object()
        token = auth_hooks.generate_file_token(user, file)
        url = urljoin(settings.BASE_URL or '/',
                      reverse('api_apella-files-downloadfile', args=(pk,)))
        response['X-File-Location'] = "%s?token=%s" % (url, token)
        return response

    @detail_route(methods=['get'], url_path='downloadfile')
    def download_get(self, request, pk=None):
        token = request.GET.get('token', None)
        if token is None:
            raise PermissionDenied("no.token")
            # url = reverse('apella-files-download', args=(pk,))
            # ui_url = getattr(settings, 'DOWNLOAD_FILE_URL', '')
            # ui_download_url = '%s?#download=%s' % (ui_url, url)
            # return HttpResponseRedirect(ui_download_url)

        file_id = auth_hooks.consume_file_token(token)
        if not file_id == int(pk):
            raise Http404

        file = get_object_or_404(ApellaFile, id=file_id)
        filename = file.file_name
        if isinstance(filename, unicode):
            filename = filename.encode('utf-8')
        filename = filename.replace('"', '')
        disp = 'attachment; filename="%s"' % filename
        if USE_X_SEND_FILE:
            response['X-Sendfile'] = file.file_content.path
        else:
            chunk_size = 8192
            response = StreamingHttpResponse(
                           FileWrapper(
                               open(file.file_content.path, 'rb'), chunk_size),
                                  content_type="application/force-download")
        response['Content-Disposition'] = disp
        return response

    def destroy(self, request, pk=None):
        obj = self.get_object()
        os_remove_file_and_log(obj, request.user.username)
        if obj.file_kind == 'foreign_diploma':
            fdd = obj.foreign_diploma_details
            if fdd:
                for extra_file in fdd.foreign_diploma_files.all():
                    fdd.foreign_diploma_files.remove(extra_file)
                    os_remove_file_and_log(extra_file, request.user.username)
                    extra_file.delete()
                fdd.delete()
        elif obj.file_kind == 'foreign_diploma_extra':
            try:
                fdd = ForeignDiplomaDetails.objects.get(id=obj.source_id)
                fdd.foreign_diploma_files.remove(obj)
            except ForeignDiplomaDetails.DoesNotExist:
                logger.info(
                    "Could not get ForeignDiplomaDetails id: %i" %
                    obj.source_id)


        return super(FilesViewSet, self).destroy(request, pk)


class UploadFilesViewSet(viewsets.ModelViewSet):

    FILE_SOURCE = {
        "Professor": "profile",
        "Candidate": "profile",
        "Candidacy": "candidacy",
        "Position": "position",
        "Registry": "registry",
        "ForeignDiplomaDetails": "foreign_diploma_details",
        "DisciplineRegistry": "discipline_registry",
        "MinistryCouncilAct": "mca_files"
    }

    @detail_route(methods=['post'])
    @transaction.atomic
    def upload(self, request, pk=None):
        obj = self.get_object()
        if 'file_upload' not in request.FILES:
            return Response(status=status.HTTP_400_BAD_REQUEST)
        file_upload = request.FILES.get('file_upload', None)
        file_kind = request.data.get('file_kind', None)
        file_description = request.data.get('file_description', None)

        if file_kind not in FILE_KIND_TO_FIELD:
            return Response(status=status.HTTP_400_BAD_REQUEST)
        if not file_upload:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        field_name, many = FILE_KIND_TO_FIELD[file_kind].values()
        if field_name not in obj._meta.get_all_field_names():
            return Response(status=status.HTTP_400_BAD_REQUEST)

        owner = request.user
        if request.user.is_helpdesk():
            if isinstance(obj, ProfessorModel) or isinstance(obj, Candidate):
                owner = obj.user
            elif isinstance(obj, Candidacy):
                owner = obj.candidate
            elif isinstance(obj, Position):
                owner = obj.author.user

        uploaded_file = ApellaFile.objects.create(
                id=get_serial('apella_file'),
                owner=owner,
                file_kind=file_kind,
                source=self.FILE_SOURCE[obj.__class__.__name__],
                source_id=obj.id,
                file_content=file_upload,
                file_name=file_upload.name,
                description=file_description,
                updated_at=datetime.utcnow())

        if not many:
            setattr(obj, field_name, uploaded_file)
        else:
            many_attr = getattr(obj, field_name)
            many_attr.add(uploaded_file)
        if isinstance(obj, Candidacy):
            obj.updated_at = datetime.utcnow()
        obj.save()
        send_emails_file(obj, file_kind)

        return Response(status=status.HTTP_200_OK)


class SyncCandidacies(object):
    @detail_route(methods=['post'])
    def sync_candidacies(self, request, pk=None):
        now = datetime.utcnow()
        candidate_user = self.get_object()
        try:
            validate_candidate_files(candidate_user.user)
        except DjangoValidationError as ve:
            return Response(ve.message, status=status.HTTP_400_BAD_REQUEST)

        active_candidacies = Candidacy.objects.filter(
            candidate=candidate_user.user,
            state='posted',
            position__state='posted',
            position__ends_at__gt=now)
        for candidacy in active_candidacies:
            link_files(candidacy, candidate_user.user)
            candidacy.updated_at = now
            candidacy.save()
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def sync_candidacies_id_file(self, request, pk=None):
        now = datetime.utcnow()
        candidate_user = self.get_object()
        try:
            validate_candidate_files(candidate_user.user)
        except DjangoValidationError as ve:
            return Response(ve.message, status=status.HTTP_400_BAD_REQUEST)

        candidacies = Candidacy.objects.filter(
            candidate=candidate_user.user,
            state='posted',
            position__state__in=['posted', 'electing', 'revoked'])
        for candidacy in candidacies:
            if not candidacy.id_file:
                link_files(candidacy, candidate_user.user, id_only=True)
                candidacy.updated_at = now
                candidacy.save()
        candidate_user.synced_id_file = True
        candidate_user.save()
        return Response(request.data, status=status.HTTP_200_OK)


class CandidateProfile(object):
    @detail_route(methods=['post'])
    def request_verification(self, request, pk=None):
        candidate_user = self.get_object()
        try:
            auth_hooks.request_user_verify(candidate_user)
            candidate_user.save()
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def request_changes(self, request, pk=None):
        candidate_user = self.get_object()
        try:
            auth_hooks.request_user_changes(candidate_user)
            candidate_user.save()
            send_user_email(
                candidate_user.user,
                'apella/emails/user_profile_request_changes_subject.txt',
                'apella/emails/user_profile_request_changes_body.txt')
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def verify_user(self, request, pk=None):
        candidate_user = self.get_object()
        try:
            auth_hooks.verify_user(candidate_user)
            if isinstance(candidate_user, ProfessorModel) and not \
                candidate_user.user.shibboleth_id and \
                candidate_user.institution and \
                candidate_user.institution.has_shibboleth and not \
                candidate_user.is_academy_member:
                    candidate_user.user.can_set_academic = True
                    candidate_user.user.save()
            candidate_user.save()
            logger.info(
                'user %s verified profile %r' %
                (request.user.username, candidate_user.id))
            send_user_email(
                candidate_user.user,
                'apella/emails/user_verified_profile_subject.txt',
                'apella/emails/user_verified_profile_body.txt')
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def reject_user(self, request, pk=None):
        candidate_user = self.get_object()
        try:
            reason = None
            if 'rejected_reason' in request.data:
                reason = request.data['rejected_reason']
            auth_hooks.reject_user(candidate_user, reason=reason)
            candidate_user.save()
            logger.info(
                'user %s rejected profile %r' %
                (request.user.username, candidate_user.id))
            send_user_email(
                candidate_user.user,
                'apella/emails/user_rejected_profile_subject.txt',
                'apella/emails/user_rejected_profile_body.txt')
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def upgrade_to_professor(self, request, pk=None):
        candidate_user = self.get_object()
        if not candidate_user.is_candidate():
            return Response(
                'not.a.candidate', status=status.HTTP_400_BAD_REQUEST)

        department = request.data.get('department', None)
        rank = request.data.get('rank', None)
        fek = request.data.get('fek', None)
        discipline_text = request.data.get('discipline_text', None)
        discipline_in_fek = request.data.get('discipline_in_fek', None)

        try:
            upgrade_candidate_to_professor(
                candidate_user.user,
                department=department,
                rank=rank,
                fek=fek,
                discipline_text=discipline_text,
                discipline_in_fek=discipline_in_fek)
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        return Response(request.data, status=status.HTTP_200_OK)


class PositionsPortal(object):
    def get_queryset(self):
        now = datetime.utcnow()
        return self.queryset.filter(
            is_historical=False,
            state='posted',
            ends_at__gte=now)


class UserApplicationMixin(object):
    def _can_accept_application(self):
        application = self.get_object()
        if application.department.institution.is_deprecated:
            return False
        apps = UserApplication.objects.filter(
                user=application.user,
                app_type=application.app_type,
                state='approved')
        if not apps:
            return True
        for app in apps:
            positions = app.position_set.all()
            for p in positions:
                if p.state != 'cancelled' and p.state != 'failed'\
                        and p.state !='successful' \
                        and not p.is_historical:
                    return False
        return True

    @detail_route(methods=['post'])
    def accept_application(self, request, pk=None):
        application = self.get_object()
        if not self._can_accept_application():
            return Response(
                'cannot.accept.application', status=status.HTTP_400_BAD_REQUEST)
        try:
            application.state = 'approved'
            application.updated_at = datetime.utcnow()
            application.save()
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        send_user_email(
            application.user,
            'apella/emails/user_application_accepted_subject.txt',
            'apella/emails/user_application_accepted_body.txt',
            {'application': application})
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def reject_application(self, request, pk=None):
        application = self.get_object()
        if application.department.institution.is_deprecated:
            return Response(
                'cannot.reject.application', status=status.HTTP_400_BAD_REQUEST)
        try:
            application.state = 'rejected'
            application.updated_at = datetime.utcnow()
            application.save()
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        send_user_email(
            application.user,
            'apella/emails/user_application_accepted_subject.txt',
            'apella/emails/user_application_rejected_body.txt',
            {'application': application})
        return Response(request.data, status=status.HTTP_200_OK)

    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user
        if user.is_anonymous():
            return UserApplication.objects.none()


        if user.is_institutionmanager():
            departments = Department.objects.filter(
                institution=user.institutionmanager.institution)
            queryset = queryset.filter(
                Q(department__in=departments) |
                Q(receiving_department__in=departments))
        elif user.is_assistant():
            departments = user.institutionmanager.departments.all()
            queryset = queryset.filter(
                Q(department__in=departments) |
                Q(receiving_department__in=departments))
        elif user.is_yyet():
            queryset = queryset.filter(
                    Q(app_type='tenure_scientist') |
                    Q(app_type='tenure_research'))
        elif user.is_professor():
            ua_ids = Position.objects.filter(
                Q(electors=user.professor) | Q(committee=user.professor)). \
                    values_list('user_application', flat=True)
            ua_ids = [ua_id for ua_id in ua_ids if ua_id]

            if user.professor.department:
                queryset = queryset.filter(
                    Q(user=user) |
                    Q(department=user.professor.department) |
                    Q(receiving_department=user.professor.department) |
                    Q(id__in=ua_ids))
            else:
                queryset = queryset.filter(Q(user=user) | Q(id__in=ua_ids))

        queryset = queryset.order_by('-id')
        return queryset


class ApellaUsers(object):
    @detail_route(methods=['post'])
    def accept_terms(self, request, pk=None):
        user = self.get_object()
        user.accepted_terms_at = datetime.utcnow()
        user.has_accepted_terms = True
        user.save()
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def release_shibboleth(self, request, pk=None):
        user = self.get_object()
        if user.login_method == 'password':
            msg = 'no.shibboleth'
            return Response(msg, status=status.HTTP_400_BAD_REQUEST)
        user.login_method = 'password'
        user.shibboleth_id = None
        user.shibboleth_idp = ''
        user.shibboleth_schac_home_organization = ''
        password = ApellaUser.objects.make_random_password(length=15)
        user.set_password(password)
        user.can_set_academic = True
        user.save()
        send_release_shibboleth_email(user, request)

        logger.info('User %s released shibboleth for %s' %
            (request.user.username, user.username))
        return Response(request.data, status=status.HTTP_200_OK)



class JiraIssues(object):
    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user
        if not user.is_helpdesk():
            queryset = queryset.filter(reporter_id=user.id)
        queryset = queryset.order_by('-updated_at')
        return queryset

class RegistryMembersSet(viewsets.ModelViewSet):

    def destroy(self, request, pk=None):
        user = self.get_object().professor.user
        department = self.get_object().registry.department
        registry_type = self.get_object().registry.type
        registry_type_el = get_registry_type_el(registry_type)

        send_user_email(
            user,
            'apella/emails/registry_remove_member_to_member_subject.txt',
            'apella/emails/registry_remove_member_to_member_body.txt',
            extra_context={
                'department_title_el': department.title.el,
                'department_title_en': department.title.en,
                'institution_title_el': department.institution.title.el,
                'institution_title_en': department.institution.title.en,
                'registry_type_el': registry_type_el,
                'registry_type_en': registry_type,
            }
        )

        return super(RegistryMembersSet, self).destroy(request, pk)


class CandidatesList(viewsets.GenericViewSet):
    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "candidates_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Departments')

        fields = [
            'κωδικός χρήστη', 'όνομα', 'επώνυμο',  'email',  'username',
            'κατάσταση πιστοποίησης', 'κατάσταση λογαριασμού',
            'αρχείο ταυτότητας/διαβατηρίου', 'αναβάθμιση', 
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        candidates = self.get_queryset()
        for c in candidates:

            profile_state = ''
            if c.is_verified:
                profile_state = 'Πιστοποιημένος'
            elif c.is_rejected:
                profile_state = 'Απορριφθείς'
            elif c.verification_pending:
                profile_state = 'Αναμονή Πιστοποίησης'
            elif not c.verification_pending and not c.is_rejected \
                    and not c.is_verified and c.changes_request:
                profile_state = 'Ζητήθηκαν αλλαγές'
            elif not c.verification_pending and not c.is_rejected \
                    and not c.is_verified and not c.changes_request:
                profile_state = 'Δεν έχει γίνει αίτηση πιστοποίησης'

            user_state = ''
            if c.user.is_active:
                user_state = 'Ενεργός'
            else:
                user_state = 'Ανενεργός'

            is_upgraded = 0
            try:
                ProfessorModel.objects.get(user__username=c.user.username)
                is_upgraded = 1
            except ProfessorModel.DoesNotExist:
                pass

            row = [
                c.user.id,
                c.user.first_name.el,
                c.user.last_name.el,
                c.user.email,
                c.user.username,
                profile_state.decode('utf-8'),
                user_state.decode('utf-8'),
                1 if c.synced_id_file else 0,
                is_upgraded,
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

class RegistryMember(object):

    @detail_route(methods=['post'])
    def toggle_registry(self, request, pk=None):
        rm = self.get_object()
        department = rm.registry.department
        professor = rm.professor
        other_type = 'internal' if rm.registry.type == 'external' \
            else 'external'
        try:
            new_reg = Registry.objects.get(department=department, type=other_type)

            registry_type = rm.registry.type
            registry_type_el = get_registry_type_el(rm.registry.type)
            send_user_email(
                professor.user,
                'apella/emails/registry_remove_member_to_member_subject.txt',
                'apella/emails/registry_remove_member_to_member_body.txt',
                extra_context={
                    'department_title_el': department.title.el,
                    'department_title_en': department.title.en,
                    'institution_title_el': department.institution.title.el,
                    'institution_title_en': department.institution.title.en,
                    'registry_type_el': registry_type_el,
                    'registry_type_en': registry_type,
                }
            )
            rm.delete()
            r = RegistryMembership.objects.create(registry=new_reg, professor=professor)
            send_user_email(
                professor.user,
                'apella/emails/registry_add_member_to_member_subject.txt',
                'apella/emails/registry_add_member_to_member_body.txt',
                extra_context={
                    'department_title_el': department.title.el,
                    'department_title_en': department.title.en,
                    'institution_title_el': department.institution.title.el,
                    'institution_title_en': department.institution.title.en,
                    'registry_type_el': get_registry_type_el(new_reg.type),
                    'registry_type_en': new_reg.type,
                }
            )
            return Response(status=status.HTTP_200_OK)

        except Registry.DoesNotExist:
            raise ValidationError(other_type+'.registry.does.not.exist')


class DisciplineRegistriesList(object):

    def get_queryset(self):
        queryset = self.queryset
        user = self.request.user
        only_active_other = self.request.query_params.get(
            'only_active_other', None)

        other_registries_active = Q(is_active=True)
        other_registries_archived = Q(deactivated_at__isnull=False)
        if user.is_assistant():
            user_departments = user.institutionmanager.departments.all()
            in_user_departments = Q(department__in=user_departments)
            q = in_user_departments | other_registries_active
            if not only_active_other:
                q |= other_registries_archived

            queryset = queryset.filter(q)

        elif user.is_institutionmanager():
            in_user_institution = Q(
                department__institution=user.institutionmanager.institution)
            q = in_user_institution | other_registries_active
            if not only_active_other:
                q |= other_registries_archived

            queryset = queryset.filter(q)

        elif user.is_professor() or user.is_candidate():
            queryset = queryset.filter(activated_at__isnull=False)


        state = self.request.query_params.get('state', None)
        if state:
            if state == 'active':
                queryset = queryset.filter(is_active=True)
            elif state == 'inactive':
                queryset = queryset.filter(
                    is_active=False,
                    activated_at__isnull=False,
                    deactivated_at__isnull=False)
            elif state == 'draft':
                queryset = queryset.filter(
                    is_active=False, activated_at__isnull=True)


        return queryset.order_by('-id')

    @detail_route(methods=['post'])
    def activate(self, request, pk=None):
        discipline_registry = self.get_object()
        try:
            validate_discipline_registry_activation(discipline_registry)
        except ValidationError as ve:
            return Response(ve.detail, status=status.HTTP_400_BAD_REQUEST)
        discipline_registry.activated_at = datetime.utcnow()
        discipline_registry.is_active = True
        discipline_registry.save()
        department = discipline_registry.department
        department.active_discipline_registry = discipline_registry
        department.save()
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def deactivate(self, request, pk=None):
        discipline_registry = self.get_object()
        department = discipline_registry.department
        department.active_discipline_registry = None
        department.save()
        discipline_registry.deactivated_at = datetime.utcnow()
        discipline_registry.is_active = False
        discipline_registry.save()
        return Response(request.data, status=status.HTTP_200_OK)

    @detail_route()
    def disciplines(self, request, pk=None):
        registry = self.get_object()
        response = HttpResponse(content_type='application/ms-excel')
        filename = "disciplines_export_" + str(registry.pk) + "_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Disciplines')

        fields = [
            'Κωδικός',
            'Γνωστικό αντικείμενο (ελ)', 'Γνωστικό αντικείμενο (en)', 
            'Επιστημονικό πεδίο (ελ)',  'Επιστημονικό πεδίο (en)', 
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        disciplines = registry.disciplines.all()
        for d in disciplines:

            row = [
                d.id,
                d.discipline_text.el,
                d.discipline_text.en,
                d.scientific_field.el,
                d.scientific_field.en,
            ]
            write_row(ws, row, i)
            i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response

    @list_route()
    def report(self, request, pk=None):
        response = HttpResponse(content_type='application/ms-excel')
        filename = "discipline_registries_export_" + \
            strftime(DATE_EXPORT_FORMAT, gmtime()) + ".xlsx"
        response['Content-Disposition'] = 'attachment; filename=' + filename

        output = StringIO.StringIO()
        wb = xlsxwriter.Workbook(
            output, {'constant_memory': True})
        ws = wb.add_worksheet('Discipline Registries')

        fields = [
            'id μητρώου',
            'id γνωστικού αντικειμένου',
            'γνωστικό αντικείμενο',
            'επιστημονικό πεδίο',
            'index_field',
            'κατάσταση μητρώου',
            'ίδρυμα',
            'τμήμα',
        ]

        k = 0
        for field in fields:
            ws.write(0, k, field.decode('utf-8'))
            k += 1

        i = 1
        registries = self.get_queryset()
        for r in registries:
            disciplines = r.disciplines.all()
            for d in disciplines:
                row = [
                    r.id,
                    d.id,
                    d.discipline_text.el,
                    d.scientific_field.el,
                    d.discipline_text_el_index,
                    get_r_state(r).decode('utf-8'),
                    r.department.institution.title.el,
                    r.department.title.el,
                ]
                write_row(ws, row, i)
                i += 1

        wb.close()
        xlsx_data = output.getvalue()
        response.write(xlsx_data)
        return response
